/*++++++++++++++
.IDENTIFICATION catutil.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Catalog Server
.KEYWORDS       CDS Catalogue Server
.VERSION  1.0   26-Dec-2006
.VERSION  1.1   19-Jun-2007: edpic; bug in match_Jname
.VERSION  1.2   15-Aug-2009: binlocaf revisited
.COMMENTS       All utilities not catalog-dependent.
---------------*/

#include <catutil.h>	/* Includes stdlib, stdio, string, ctype, math */
#include <sys/types.h>
#include <sys/stat.h>

#if 0
#include <unistd.h>
#include <fcntl.h>	/* O_BINARY */

#ifndef O_BINARY	/* Required for MS-DOS	*/
#define O_BINARY 0
#endif

#endif

/*==================================================================
		Simple transformations
 *==================================================================*/

int strloc(char *text, int c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
{
  char *s;
    for ( s = text; *s && (*s != c); s++) ;
    return(s-text);
}

char *strdup2(char *a, char *b)
/*++++++++++++++++
.PURPOSE  Create a concatenation of 2 strings
.RETURNS  a . b
-----------------*/
{
  int len; char *s;
    len = strlen(a);
    s = malloc(len+1+strlen(b));
    strcpy(s, a); strcpy(s+len, b);
    return(s);
}

int isodate(char *buf, time_t t)
/*++++++++++++++++
.PURPOSE  Edit the date in ISO format
.RETURNS  Length (19), NO ZERO byte added
-----------------*/
{
  struct tm *T;
    if (t == 0)
        t = time(0);
    T = gmtime(&t);
    if (T->tm_year<1900) T->tm_year += 1900;
    sprintf(buf, "%04d-%02d-%02dT%02d:%02d:%02dZ",
            T->tm_year, T->tm_mon+1, T->tm_mday,
            T->tm_hour, T->tm_min, T->tm_sec);
    return(strlen(buf));
}

int4 get_delay(char *str)
/*++++++++++++++++
.PURPOSE  Get a delay (s by default)
.RETURNS  The delay, in seconds / -1 if not correct
.REMARKS  Use suffix 's'=sec, 'm'=min, 'h'=hour, 'd'=day, 'w'=week
-----------------*/
{
  static char symb[] = "mhdwj";
  static int4 f2s[] = { 60, 3600, 86400, 7*86400, 86400 };
    int4 value;
    if (isdigit(*str)) {
	value = atoi(str);;
	while (isdigit(*str)) str++;
	if (islower(*str)) { int i;
	    i = strloc(symb, *str);
	    if (symb[i]) value *= f2s[i];
	}
    }
    else value = -1;
    return(value);
}

int get_bits(unsigned char *a, int b, int len)
/*++++++++++++++++
.PURPOSE  Get a value made of 'len' bits
          starting from byte position 'a', bit offset 'b'.
.RETURNS  The value
.REMARKS  Independent of the architecture.
-----------------*/
{
   static unsigned int mask[] = {
    0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f,
    0x00ff, 0x01ff, 0x03ff, 0x07ff, 0x0fff, 0x1fff, 0x3fff, 0x7fff,
    0x00ffff, 0x01ffff, 0x03ffff, 0x07ffff, 0x0fffff,
              0x1fffff, 0x3fffff, 0x7fffff,
    0x00ffffff, 0x01ffffff, 0x03ffffff, 0x07ffffff, 0x0fffffff,
                0x1fffffff, 0x3fffffff, 0x7fffffff,
    0xffffffff } ;
  int value;
  unsigned char *ac;
  int nb ;      /* remaining bits to get */

    ac = a + (b>>3);    /* Byte position  */
    value = *(ac++);    /* Initialisation */
    nb = len;           /* Bits to read.  */
    nb -= 8 - (b&7);    /* Useful bits    */

    /* We don't care about the leftmost bits,
       these will be removed at the end
       (we assume that len <= 32...)
    */

    while (nb >= 8) { value = (value<<8) | *(ac++); nb -= 8; }
    if (nb < 0)         /* I've read too much */
        value >>= (-nb);
    else if (nb)        /* Remainder bits  */
        value = (value<<nb) | (*ac >> (8-nb));

    return(value&mask[len]);
}

void verify_swap()
/*++++++++++++++++
.PURPOSE  Verify the swap is OK for this architecture
.RETURNS  exit if error
-----------------*/
{
  static int value = 0x010203;
  char *v;
    v = (char *)(&value);
#if SWAPPING == 0
    if ((v[0] == 0) && (v[1] == 1) && (v[2] == 2))
	return;
#endif
#if SWAPPING == 1
    if ((v[0] == 1) && (v[1] == 0) && (v[2] == 3))
	return;
#endif
#if SWAPPING == 2
    if ((v[0] == 3) && (v[1] == 2) && (v[2] == 1))
	return;
#endif
    fprintf(stderr, "****verify_swap: BAD SWAPPING (%d)\n", SWAPPING);
    exit(2);
}

void swap_a1(unsigned char *a, int size)
/*++++++++++++++++
.PURPOSE  Swap in memory short int's
.RETURNS  ---
.REMARKS  size must be a multiple of 2 !
-----------------*/
{
  register unsigned char *p, *e;
  register int c;
    for (p=a, e=p+size; p<e; p += 2) {
	c = p[0]; p[0] = p[1]; p[1] = c; 
    }
}

void swap_a2(unsigned char *a, int size)
/*++++++++++++++++
.PURPOSE  Swap in memory 2, 4, 8, or 16 bytes
.RETURNS  ---
.REMARKS  size should be a power of 2. Exchange right/left
-----------------*/
{
  register int i, j, c;
    for (i=0, j=size-1; i<j; ++i, j--) {
	c = a[i];
	a[i] = a[j];
	a[j] = c; 
    }
}

void swap_a3(unsigned char *a, int size)
/*++++++++++++++++
.PURPOSE  Swap in memory an array of 4-byte words (exchange 0/3 and 1/2
.RETURNS  ---
.REMARKS  size must be a multiple of 4!
-----------------*/
{
  register unsigned char *p, *e;
  register int c;
    for (p=a, e=p+size; p<e; p += 4) {
	c = p[0]; p[0] = p[3]; p[3] = c;
	c = p[1]; p[1] = p[2]; p[2] = c;
    }
}

void swap_a7(unsigned char *a, int size)
/*++++++++++++++++
.PURPOSE  Swap in memory an array of 8-byte words (exchange 0/7, 1/6, etc
.RETURNS  ---
.REMARKS  size must be a multiple of 4!
-----------------*/
{
  register unsigned char *p, *e;
  register int c;
    for (p=a, e=p+size; p<e; p += 8) {
	c = p[0]; p[0] = p[7]; p[7] = c;
	c = p[1]; p[1] = p[6]; p[6] = c;
	c = p[2]; p[2] = p[5]; p[5] = c;
	c = p[3]; p[3] = p[4]; p[4] = c;
    }
}

int isqrt(int n)
/*++++++++++++++++
.PURPOSE Compute the square root (integer s such that s*s <= n)
.RETURNS Integer root
.REMARKS Algorithm
   \begin{TeX}
   {x}_{k+1} = \frac{1}{2}\left(x_k + \frac{ n }{x}_{k}\right), \ k \ge 0.
   \end{TeX}
-----------------*/
{
  int x0, x1, d;
  unsigned int m;
    /* First approximation = halve bits */
    if (n<=1) return(n);
    for (m=0xc0000000; (n&m)==0; m>>=2) ;
    for (x0=0; m; m>>=1) x0++;
    x0 = 1+(n>>x0); /* x0 = (n+1)/2; */
    d = 2;
    while (d!=0) {
	x1 = (x0+(n/x0))/2;
	d = x1 - x0;
	x0 = x1;
	if ((d == 1) && (x1*x1 > n)) break;
    }
    return(x0);
}

int ihypot(int a, int b)
/*++++++++++++++++
.PURPOSE Compute the square root (integer s such that s*s <= n)
.RETURNS Integer root
.REMARKS Algorithm
   \begin{TeX}
   {x}_{k+1} = \frac{1}{2}\left(x_k + \frac{ n }{x}_{k}\right), \ k \ge 0.
   \end{TeX}
-----------------*/
{
  int8 n;
  int8 x0, x1, d;
    n = (int8)a*a + (int8)b*b;
    if (n<=1) return(n);
    /* First approximation: about (|a|+|b|)/sqrt(2) */
    x0 = (a>=0? a : -a) + (b>=0 ? b : -b);
    x0 = x0*5/7; 
    d = 2;
    while (d!=0) {
	x1 = (x0+(n/x0))/2;
	d = x1 - x0;
	if ((d == 1) && (x1*x1 > n)) break;
	x0 = x1;
    }
    return((int)x0);
}

int strmatch(char *template, char *name)
/*++++++++++++++++
.PURPOSE  Check name against template
.RETURNS  1 (Match) / 0 (mismatch)
.REMARKS  * ? [...] recognized as wild chars
-----------------*/
{
  char *t, *f; int m, not;
    for (t = template, f = name; *t; t++) switch(*t) {
      case '?':                 /* Matches one character */
        if (*f) f++;
        else return(0);
        continue;
      case '*':                 /* Match 0 or n characters */
        if (! *++t) return(1);  /* The * is the last char of template */
#if 1           /* Try LONGEST possible matching */
        for (m = strlen(f); --m >= 0; ) {
            if (strmatch(t, f+m))
                return(1);
        }
#else           /* Try SHORTEST possible matching */
        for (; *f; f++) {       /* Look for first matching char */
            if (isalnum(*t) && (*f != *t)) continue;
            if (strmatch(t,f))  /* Recursive call */
                return(1);
        }
#endif
        return(0);
      case '[':         /* A set of characters is provided */
        ++t; m=not=0;
        if (*t == '^') { not = 1 ; t++ ; }
        if (*t == ']') { /* Closing brace is acce[table */
            t++;
            m = (*f == ']');
        }
        while ((m==0) && *t) {  /* Check: does *f be in set ?   */
            if (*t == ']') return(0);           /* End of set   */
            if ((t[1] == '-') && (t[2] != ']')) {      /* Range */
                m = (*f >= t[0]) && (*f <= t[2]);
                t += 3;
            }
            else m = (*f == *t), t++;
        }
        m ^= not;       /* V4.9 */
        if (m == 0) return(0);
        t += strloc(t, ']');
        f++;
        continue;
      case '\\':        /* Escaped Char. NO BREAK HERE! */
        t++;
      default:
        if (*t != *f) return(0);
        f++;
        continue;
    }
    return((!*t) && (!*f));     /* Returns 1 if *t AND *f are NUL */
}

/*==================================================================
		Retrieve data for a sorted array
 *==================================================================*/

int binloc(int4 *start, int items, int4 value)
/*++++++++++++++++
.PURPOSE  Find index in table such that    start[i] <= value < start[i+1]
.RETURNS  index / -1 (< *start) / items (>= start[items])
-----------------*/
{
  int4 *low, *high, *med ;
    if (value < *start) return(-1) ;
    low=start, high=start+items-1;
    if (value >= *high) low = high;
    while ((high-low) > 1) {
        med = low + (high-low)/2 ;
        if (value < *med) high = med;
        else low = med ;
    }
    return(low - start) ;
}

int binlocf(char *start, int bytes, int reclen, int4 (*f)(), int4 value)
/*++++++++++++++++
.PURPOSE  Find index in table such that   f(p) <= value < f(p+reclen)
.RETURNS  Byte position in start / -1 / bytes
-----------------*/
{
  char *low, *high, *med ;
  int4 ival;
    ival = (*f)(start);
    if (value < ival) return(-1);    /* Value too small */
    if (value == ival) return(0);    /* Coindices...    */
    low=start, high=start+bytes-reclen;
    ival = (*f)(high);
    if (value > ival) return(bytes); /* Value too large */
    if (value == ival) {             /* just ok for top */
        for (med = high-reclen; (*f)(med) == value; med -= reclen) ;
        return(reclen + (med-start));
    }
    while ((high-low) > reclen) {
        med = low + reclen*((high-low)/reclen/2) ;
        if (value < (*f)(med)) high = med;
        else low = med ;
    }
    return(low - start) ;
}

int abinloc(void **start, int items, void **new, SORT_FCT cmp)
/*++++++++++++++++
.PURPOSE  Find index in table of pointers such that
          start[i] <= new < start[i+1]
.RETURNS  Index in start (-1 = before first, items=after last)
-----------------*/
{
  void **low, **high, **med ;
    if ((*cmp)(new, start) < 0) return(-1) ;
    low=start, high=start+items-1;
    if ((*cmp)(new, high) >= 0) low = high;
    while ((high-low) > 1) {
        med = low + ((high-low)/2) ;
        if ((*cmp)(new, med) < 0) high = med;
        else low = med ;
    }
    return(low - start) ;
}

/*==================================================================
		dealing with soft links
 *==================================================================*/
char *realdir(char *path)
/*+++++++++++
.PURPOSE Find out the 'real' directory of a file
.RETURNS A new piece of memory with real directory name
.REMARKS Returned directory is ALWAYS TERMiNATED BY A /
	To find out the physical directory of a program, use
	realdir(getenv("_"));
------------*/
{
  char *name, *astring=0, wd[256];
  struct stat sbuf;
  int st, nl;

    for (nl=0, name=path; nl<64; nl++) {
        st = lstat(name, &sbuf);
        if (st) {
            fprintf(stderr, "#+++realdir(%s): ", path);
            perror(name);
            return((char *)0);
        }
        if (S_ISLNK(sbuf.st_mode)) { char *tofree;
            tofree = astring;
            astring = malloc(sbuf.st_size+3);
            st = readlink(name, astring, sbuf.st_size+1);
            if (st>=0) astring[st]=0;
            else {
                fprintf(stderr, "#+++readlink(%s): ", name);
                perror("");
                if (tofree) free(tofree);
                free(astring);
                return((char *)0);
            }
            /* If relative path: prepend the directory name */
            if (astring[0] != '/') {
                char *rname = astring;
                if (name[0] == '/') {    /* Link relative to name */
                    astring = malloc(strlen(name) + strlen(rname) + 3);
                    strcpy(astring, name);
                    strcpy(strrchr(astring,'/')+1, rname);
                }
                else {  /* Link relative to current dir */
                    if (!getcwd(wd, sizeof(wd))) {
                        fprintf(stderr, "#+++readlink(%s): ", path);
                        perror("getcwd");
                        free(astring);
                        return((char *)0);
                    }
                    astring = malloc(strlen(wd) + strlen(rname) + 3);
                    strcpy(astring,wd);
                    strcat(astring, "/");
                    strcat(astring, rname);
                }
                free(rname);
            }
            if (tofree) free(tofree);
            name = astring;
        }
        else break;
    }
    if (nl>=64) {
        fprintf(stderr, "#+++realdir(%s): link loop!\n", path);
        free(astring);
        return((char *)0);
    }

    /* Convert relative path to absolute */
    if (name == path) {
	if (*path != '/') {
	    if (!getcwd(wd, sizeof(wd))) {
		fprintf(stderr, "#+++readlink(%s): ", path);
		perror("getcwd");
		return((char *)0);
	    }
	    st = strlen(wd);
	    name = malloc(st+1+strlen(path));
	    strcpy(name, wd);
	    name[st++] = '/';
	    strcpy(name+st, path);
	}
	else {
            name = malloc(strlen(path)+3);
            strcpy(name, path);
	}
    }
    /*fprintf(stderr, "#...readdir(%s): found '%s'\n", path, name);*/
    if ((astring = strrchr(name, '/')))
        astring[1] = 0;
    else strcpy(name, "./");
    return(name);
}

char *progdir(char *argv0)
/*+++++++++++
.PURPOSE Find out the 'real' directory of a program
.RETURNS A new piece of memory with real directory name
.REMARKS Returned directory is ALWAYS TERMiNATED BY A /
------------*/
{
  char *path;
    if (argv0[0] == '/') path = (char *)0;
    else path = getenv("_");
    if (!path) path = argv0;
    return(realdir(path));
}

/*==================================================================
		Interpretation of strings
 *==================================================================*/

int match_int4 (char *string, int4 *value)
/*++++++++++++++++
.PURPOSE  Analyze the incoming and get integer number
.RETURNS  How many bytes matched
.REMARKS  If no digit ==> set to nan4;
-----------------*/
{
  char *p; int4 val;
    if (!string) return(-1);
    while (*string == ' ') string++;
    p = string;
    if ((*p == '+') || (*p == '-')) p++;
    if (isdigit(*p)) {
        val = 0;
        while (isdigit(*p)) val = val*10 + (*(p++)&15);
	if (*string == '-') val = -val;
	*value = val;
    }
    else if (*string == '-') {		/* Accept minuses as NaN */
	while (*p == '-') p++;
	*value = nan4;
    }
    else return(0);			/* Nothing found! */
    return(p-string);
}

int match_unum (char *string, double *value, int *ndec)
/*++++++++++++++++
.PURPOSE  Analyze the incoming and get unsigned number (no E notation)
	  (accept e notation)
.RETURNS  How many bytes matched
.REMARKS  Initial blanks return 0. On return, ndec = number of decimals
	  -1 (no decimal point, i.e. integer value)
	  Value not changed if no match (returned number = 0)
-----------------*/
{
  char *p, *adot;
    if (!string) return(-1);
    for (p=string; isdigit(*p); p++) ;
    adot = p+1;
    if (*p == '.') for ( ++p; isdigit(*p); p++) ;
    if (*p == 'e') for (p+=2; isdigit(*p); p++) ;
    if (p > string) *value = atof(string) ;
    if (ndec) *ndec = p - adot;
    return(p-string);
}

int match_num (char *string, double *value, int *ndec)
/*++++++++++++++++
.PURPOSE  Analyze the incoming and get signed number (accept 'e' notation)
.RETURNS  How many bytes matched
-----------------*/
{
  char *p; 
    if (!string) return(-1);
    while (*string == ' ') string++;
    p = string ;
    if ((*p == '+') || (*p == '-')) p++ ;
    if (isdigit(*p)) {
        p += match_unum(p, value, ndec) ;
        if (*string == '-') *value = -*value ;
    }
    else return(0);
    return(p-string);
}

int match_num2 (char *string, double values[2])
/*++++++++++++++++
.PURPOSE  Find out 2 numbers separated by , / : or ..
.RETURNS  Number of bytes matched
.REMARKS  values set to NaN or +/-Inf if absent
-----------------*/
{
  int n;
  char *p, *s, *s2;

    if (!string) return(-1);
    /* Gobble blanks */
    for (s=string; *s== ' '; s++) ;
    p = s; 

    /* Get leftmost number */
    n = match_num(p, values, (int *)0) ;
    if (n == 0) values[0] = 0./0.;
    else        p += n;
    while (*p == ' ') p++;
    s = p;

    /* Skip the separator */
    s2 = p;
    if ((*p == ',') || (*p == ':') || (*p == '/') || (*p == 'x'))
	p++;
    else if (p[0] == '.') {
	if (p[1] == '.') p+=2;
	else if (p[-1] == '.') p++;
    }
    if (p == s2)	/* No separator -- value#2 is NaN */
	values[1] = 0./0.;
    else {	/* A separator is matched. Missing values are +/-Inf */
	if (values[0] == values[0]) ;	/* A true number, OK */
	else values[0] = -1./0.;	/* -Infinity */
        n = match_num(p, values+1, (int *)0) ;
        if (n == 0) values[1] = 1./0.;	/* +Infinity */
        else        p += n;
    }

    return(p-string) ;
}

int match_Jname (char *arg, int4 *mas, int4 *Jprec)
/*++++++++++++++++
.PURPOSE  Interpret a string for a Jname
.RETURNS  Number of bytes matched
.REMARKS  Results of RA and Dec in mas. 
	  Value of last decimals saved in Jprec, in mas
	  (beware, Jprec negative when declination negative!)
-----------------*/
{
  static int step1[] = {	/* Factors to convert to ms */
    36000000,			/* # (deg)	*/
    3600000, 360000,		/* HH or DD DDd */
    60000, 6000,		/* HHMM  HHMMm  */
    1000, 100,			/* HHMMSS[s]    */
    10, 1			/* HHMMSSss[s]  */
  };
  char apart[12], sign, dpart[11], *p;
  int a, d, i, ms;

    if (!arg) return(-1);
    p = arg;
    if (toupper(*p) == 'J') p++;
    for (i=a=0; (i<10) && p[i]; i++) {
        if (p[i] == '.') {
            if (a==6) continue;
            fprintf(stderr, "#***Bad Jname(RA) : %s\n", arg);
            return(-1);
        }
        if (isdigit(p[i])) apart[a++] = p[i];
        else break;
    }
    apart[a] = 0;
    sign = p[i++];
    if ((sign != '+') && (sign != '-')) {
        fprintf(stderr, "#***Bad Jname(sgn): %s\n", arg);
        return(-1);
    }
    p += i;
    for (d=i=0; (i<10) && p[i]; i++) {
        if (p[i] == '.') {
            if (d==6) continue;
            fprintf(stderr, "#***Bad Jname(Dec): %s\n", arg);
            return(-1);
        }
        if (isdigit(p[i])) dpart[d++] = p[i];
        else break;
    }
    dpart[d] = 0;
    if ((a<4)||(d<2)) {
        fprintf(stderr, "#***Bad Jname(...): %s\n", arg);
        return(-1);
    }
    p += i;			/* Modified V1.1 */

    /* Convert to 1/1000" */
    step1[2] = a == 3 ? 360000 : 600000;
    step1[4] = a == 5 ?   6000 :  10000;
    for (i=ms=0; apart[i]; i++) ms += (apart[i]-'0')*step1[i];
    Jprec[0] = step1[a-1]*15;
    mas[0] = ms*15;

    step1[2] = d == 3 ? 360000 : 600000;
    step1[4] = d == 5 ?   6000 :  10000;
    for (i=ms=0; dpart[i]; i++) ms += (dpart[i]-'0')*step1[i];
    Jprec[1] = step1[d-1];
    mas[1] = ms; 
    if (sign == '-') {
	mas[1] = -ms;
	Jprec[1] = -Jprec[1];
    }

    return(p-arg);
}

int match_Gname (char *arg, int4 *udeg, int4 *prec)
/*++++++++++++++++
.PURPOSE  Interpret a string for a Gname
.RETURNS  Number of bytes matched
.REMARKS  Results of Glon and Glat in udeg.
	  Value of last decimals saved in prec, in udeg
	  (beware, prec negative when declination negative!)
-----------------*/
{
  double value;
  char sign, *p;
  int ndec, i;

    if (!arg) return(-1);
    p = arg;
    if (toupper(*p) == 'G') p++;
    i = match_num(p, &value, &ndec);
    if (value<0) value += 360.;
    udeg[0] = (value+1.e-12) * 1.e6;
    for (value=1000000; ndec>0; ndec--) value /= 10; prec[0] = value;

    /* Latitude part */
    p += i;
    sign = *(p++); 
    if ((sign != '+') && (sign != '-')) {
        fprintf(stderr, "#***Bad Gname(sgn): %s\n", arg);
        return(-1);
    }
    i = match_unum(p, &value, &ndec);
    udeg[1] = (value+1.e-12) * 1.e6;
    for (value=1000000; ndec>0; ndec--) value /= 10; prec[1] = value;
    if (sign == '-') {
	prec[1] = -prec[1];
	udeg[1] = -udeg[1];
    }
    
    p += i;
    return(p-arg);
}

int match_center (char *string, double center[2])
/*++++++++++++++++
.PURPOSE  Interpret a string for RA + DEC
.RETURNS  Number of bytes matched
.REMARKS  Value 999. if not found. Accept jhhmm...
-----------------*/
{
  double v, factor ;
  char *p, *s, *w[6];
  int ndec, nw;

    factor = 1;
    center[0] = center[1] = 999. ;
    if (!string) return(-1);
    p = string ; while (*p == ' ') p++ ;        /* Skip leading blanks */
    if (isalpha(*p)) {				/* Accept Jname of Gname */
      int4 icoo[2], prec[2]; int bytes; double factor;
	bytes = toupper(*p);
	if (bytes == 'J') {
	    bytes = match_Jname(p, icoo, prec);
	    factor = 3.6e6; 
	}
	else if (bytes == 'G') {
	    bytes = match_Gname(p, icoo, prec);
	    factor = 1.e6;
	}
	else return(-1);
        if (bytes < 0) return(bytes);
        bytes += (p-string);
#if 0
        /* Mean value of truncation effect = 45/100 */
        icoo[0] += (45*prec[0])/100;
        icoo[1] += (45*prec[1])/100;		/* Jprec[1 <0 if mas[1]<0 */
#endif
        center[0] = icoo[0]; center[1] = icoo[1];
        center[0] /= factor; center[1] /= factor;
        return(bytes);
    }
    p += match_unum(p, center, &ndec) ;         /* Get first number */
    if (*p == ' ') {    /* Blanks embedded. Could be sexagesimal... */
        while (*p == ' ') p++ ;
        if (ndec<0) {
            /* We count the number of words.
               Is a sign is embedded, the delimiter RA/Dec is clear.
               Otherwise, we introduce a '+' sign at the relevant position.
            */
            for (nw = 1, s = p-1; *s; s++) {
                if ((*s == '+') || (*s == '-')) break;
                if (*s == ' ') {
                    for (++s; *s == ' '; s++) ;
                    if (*s) w[nw++] = --s;
                    if (nw >= 6) break;
                }
            }
            if (!*s) {          /* No sign found. Introduce it. */
                nw -= (nw/2);   /* move backwards ... */
                s = w[nw];
                if (*s == ' ') *s = '+';
            }
        }
    }
    /* As soon as we find a decimal point, we can't be in RA any more */
    if ((*p == ',')||(*p == '\t'))	/* Now in Dec part... */
	p++;
    else while (ndec < 0) {
        if ((*p == ',')||(*p == '\t')){ /* Starting Dec part. */
            p++;
            break;
        }
        if (*p == ':') p++;
        if (!isdigit(*p)) break;
        factor /= 60.;
        if (factor < 1./3600.) break;
        p += match_unum(p, &v, &ndec);
        while (*p == ' ') p++ ;
        center[0] += factor*v;
    }
    if (factor != 1.) center[0] *= 15;  /* sexagesimal was found */
    while (*p == ' ') p++;
    s = p ;     /* Normally, the sign */
    if ((*s == '+') || (*s == '-')) p++ ;
    factor = 1;
    for (ndec = -1; ndec < 0; factor /= 60.) {
        if (*p == ':') p++ ;
        if (!isdigit(*p)) break;
        p += match_unum(p, &v, &ndec);
        while (*p == ' ') p++ ;
        if (factor == 1.) center[1] = v;
        else center[1] += factor*v ;
    }
    if (*s == '-') center[1] = -center[1] ;

    return(p-string) ;
}

/*==================================================================
		Get limits (range)
 *==================================================================*/

int get_lims(char *string, double values[2])
/*++++++++++++++++
.PURPOSE  Find out the limits (2 numbers separated by ,)
.RETURNS  0 / 1 (1 number)/ 2 (right number) / 3 (full range) / -1=error
.REMARKS  Accept p=min1,max1=min2,max2 ==> sqrt(min1^2+min2^2)
-----------------*/
{
  int found=0;
  int n;

    if (!string) return(-1);
    n = match_num2(string, values);
    if (string[n]) {	/* Not totally matched ? */
	/* fprintf(stderr, "#***'%s' not understood in limits: %s\n",
		string+n, string); */
	return(-1);
    }
    found =  values[0] == values[0];	/* a true number found*/
    if      (values[1] == values[1]) 	/* a true number .... */
	found |= 2;

    return(found);
}

int get_lim4(char *string, int4 lim[2], double factor)
/*++++++++++++++++
.PURPOSE  Find out the limits (2 numbers separated by ,) as integer numbers
.RETURNS  0 (assume nan) / 1 / 2 / 3 / -1=error
.REMARKS  NaN ==> nan4, Infinity ==> 2^31-1
-----------------*/
{
  int found;
  double dval[2];

    found = get_lims(string, dval);
    if (found <= 0) {		/* Nothing, assume nan */
	lim[0] = lim[1] = nan4;
	return(found);
    }
    if (found == 1) dval[1] = dval[0];
    dval[0] *= factor; dval[1] *= factor;
    lim[0] = dval[0] < -2147483647.0 ? -2147483647 : dval[0];
    lim[1] = dval[1] >  2147483647.0 ?  2147483647 : dval[1];
    return(found);
}

/*==================================================================
		Edition of numbers
 *==================================================================*/

static int p10[] = { 1, 10, 100, 1000, 10000, 100000, 1000000, 10000000 };
static int r10[] = {  0, 4,  49,  499,  4999,  49999,  499999,  4999999 };

int edpic (char *buf, char *pic, int4 val)
/*++++++++++++++++
.PURPOSE  Edit a value according to a 'picture'
.RETURNS  Number of bytes edited (strlen(pic))
.REMARKS  '+' is sign, '0' is blank or digit
	NO TERMINATING ZERO
-----------------*/
{
  int4 v, base;
  int len, i, s;
    len = i = strlen(pic);
    v = val;
    if (v<0) { v = -v; s = '-'; } 
    else s = strchr(pic, '+') ? '+' : ' ';
    while (--i>=0) {
	if (isdigit(pic[i])) 
	     base = pic[i] - ('0'-1) ;
	else if (pic[i] == '+') 
	    base = 1;
	else { buf[i] = pic[i]; continue; }
	if (base == 1) {	/* '0' */
	    if (v==0) { buf[i] = s; s = ' '; continue; }
	    base = 10;
	}
	buf[i] = '0' | (v%base);
	v /= base;
    }

    /* Verify the number not too large */
    if (v || (s != ' ')) {
	fprintf(stderr, "#***edpic(%s, %d): value does not fit\n", pic, val);
	for (i=len; --i>=0; ) buf[i] = '*';
    }
    return(len);
}

int edsex_ra (char *buf, int4 ramas, int ndec)
/*++++++++++++++++
.PURPOSE  Edit RA in sexagesimal (HH:MM:SS...)
.RETURNS  Number of bytes edited (9+ndec)
.REMARKS  Value rounded. Note that the result might be 24:00:00.
	ndec must be in range 0..4
-----------------*/
{
  int i; int4 value;
    value = ramas;
    value = value - (value/3);	/* Value in 0.1ms */

    /* Edition of fractional part */
    if (ndec>0) {
	if (ndec>4) ndec=4;
	value = (value+r10[4-ndec])/p10[4-ndec];
        for (i=9+ndec; --i>8; ) { 
	    buf[i] = '0' + value%10; 
	    value /= 10; 
	}
        buf[8] = '.' ;
    }
    else {
	ndec = -1;
	value = (value+r10[4])/p10[4];
    }

    /* Edition of HH:MM:SS */
    buf[7] = '0' + value%10, value /= 10 ;
    buf[6] = '0' + value%6,  value /= 6 ;
    buf[5] = ':' ;
    buf[4] = '0' + value%10, value /= 10 ;
    buf[3] = '0' + value%6,  value /= 6 ;
    buf[2] = ':' ;
    buf[1] = '0' + value%10, value /= 10 ;
    buf[0] = '0' + value ;
    return(ndec+9);
}

int eddeg_ra (char *buf, int4 ramas, int ndec)
/*++++++++++++++++
.PURPOSE  Edit RA in decimal degrees
.RETURNS  Number of bytes edited (4+ndec)
.REMARKS  Value rounded. Note that the result might be 360.0000
	ndec must be in range 0..7
	Remark: 10^7/3600000 = 3 * (1 - 2/27)
-----------------*/
{
  int i, deg; int4 value;
    deg   = ramas/3600000;	/* Integer degree part ...    */
    value = 3*(ramas%3600000);	/* Degree fraction, in 1/3mas */
    value -= ((2*value+13)/27);	/* Degree fraction in 10-7deg */

    if (ndec>0) {
	if (ndec>7) ndec=7;
	value = (value+r10[7-ndec])/p10[7-ndec];
	if (value >= p10[ndec]) { deg++; value -= p10[ndec]; }
        for (i=4+ndec; --i>3; ) { 
	    buf[i] = '0' + value%10; 
	    value /= 10; 
	}
        buf[3] = '.' ;
    }
    else {
	ndec = -1;
	if ((value + r10[7]) >= p10[7]) deg++;
    }
    buf[2] = '0' + deg%10, deg /= 10 ;
    buf[1] = '0' + deg%10, deg /= 10 ;
    buf[0] = '0' + deg ;
    return(ndec+4);
}

int edmas_ra (char *buf, int4 ramas)
/*++++++++++++++++
.PURPOSE  Edit RA in mas
.RETURNS  Number of bytes edited (10)
.REMARKS  Filled with blanks
-----------------*/
{
  int i; int4 value;
    value = ramas;
    i = 10;
    buf[--i] = (value%10) + '0';
    value /= 10;
    while (value>0) {
	buf[--i] = (value%10) + '0';
	value /= 10;
    }
    while (i>0) buf[--i] = ' ';
    return(10);
}

int edsex_de (char *buf, int4 demas, int ndec)
/*++++++++++++++++
.PURPOSE  Edit a Decl (unit=1mas) into +DD:MM:SS.SSS
.RETURNS  Number of bytes edited (10+ndec)
.REMARKS  Value rounded. 
	ndec must be in range 0..3
-----------------*/
{
  int i; int4 value;
    value = demas;
    if (value >= 0) buf[0] = '+';
    else {          buf[0] = '-'; value = -value ; }

    if (ndec>0) {
	if (ndec>3) ndec=3;
        value = (value+r10[3-ndec])/p10[3-ndec];
	for (i=10+ndec; --i>9;  ) { 
	    buf[i] = '0' + value%10; 
	    value /= 10; 
	}
        buf[9] = '.' ;
    }
    else {
	ndec = -1;
	value = (value+r10[3])/p10[3];
    }
    buf[8] = '0' + value%10, value /= 10 ;
    buf[7] = '0' + value%6,  value /= 6 ;
    buf[6] = ':' ;
    buf[5] = '0' + value%10, value /= 10 ;
    buf[4] = '0' + value%6,  value /= 6 ;
    buf[3] = ':' ;
    buf[2] = '0' + value%10, value /= 10 ;
    buf[1] = '0' + value ;
    return(ndec+10) ;
}

int eddeg_de (char *buf, int4 demas, int ndec)
/*++++++++++++++++
.PURPOSE  Edit a Decl (unit=1mas) into +DD.ddddd (decimal degrees)
.RETURNS  Number of bytes edited (4+ndec)
.REMARKS  Value rounded. 
	ndec must be in range 0..3
-----------------*/
{
  int i, deg; int4 value;
    value = demas;
    if (value >= 0) buf[0] = '+';
    else {          buf[0] = '-'; value = -value ; }
    deg   = value/3600000;	/* Integer degree part ...    */
    value = 3*(value%3600000);	/* Degree fraction, in 1/3mas */
    value -= ((2*value+13)/27);	/* Degree fraction in 10-7deg */

    if (ndec>0) {
	if (ndec>7) ndec=7;
	value = (value+r10[7-ndec])/p10[7-ndec];
	if (value >= p10[ndec]) { deg++; value -= p10[ndec]; }
        for (i=4+ndec; --i>3; ) { 
	    buf[i] = '0' + value%10; 
	    value /= 10; 
	}
        buf[3] = '.' ;
    }
    else {
	ndec = -1;
	if ((value + r10[7]) >= p10[7]) deg++;
    }
    buf[2] = '0' + deg%10, deg /= 10 ;
    buf[1] = '0' + deg;
    return(ndec+4);
}

int edmas_de (char *buf, int4 demas)
/*++++++++++++++++
.PURPOSE  Edit Dec in mas
.RETURNS  Number of bytes edited (10)
.REMARKS  Filled with zeroes
-----------------*/
{
  int i; int4 value;
    if (demas >= 0) { buf[0] = '+'; value =  demas; }
    else            { buf[0] = '-'; value = -demas; }
    for (i=10; --i>0; ) {
	buf[i] = (value%10) + '0';
	value /= 10;
    }
    return(10);
}

/*==================================================================
		Date/Time Transformation
 *==================================================================*/

struct tm *jd2tm(double JD, struct tm *T)
/*++++++
.PURPOSE Transform Julian Date to time structure
.RETURNS Time as a structure
.REMARKS  INPUT:        Julian date
------*/
{
  static struct tm tm_def;
  register int4 j, n4, nd10;
  double x;

    if (!T) T = &tm_def;
    x = JD + 0.5e0 + 0.5e0/86400.e0;    /* Round day    */
    j = x;
    if (x < 0)   --j;
    x = x - j;                          /* Fraction of day      */

    T->tm_year = -4712;
    while (j <= 4480)   { j += 146097L; T->tm_year -= 400;}

    n4 = 4*(j+((2*((4*j-17918L)/146097L)*3)/4+1)/2-37);
    nd10=10*( ((n4-237)%1461)/4)+5;

    T->tm_year += n4/1461;
    T->tm_mon  = (nd10/306+2)%12;
    T->tm_mday = (nd10%306)/10+1;

    x *= 86400.e0;                      /* Seconds      */
    j = x;
    T->tm_hour = j/3600;
    T->tm_min  = 0;
    T->tm_sec  = j%3600;

    return(T);
}

void jd2date(double JD, char date[11])
/*++++++
.PURPOSE Transform Julian Date to time date string
.RETURNS Time as a string in date[]
.REMARKS INPUT: Julian date
------*/
{
  struct tm T;
  static double JDp = -1.e12;
  static char datep[12];

  if (JD != JDp) {
      jd2tm(JD, &T);
      sprintf(datep, "%04d-%02d-%02d", T.tm_year, T.tm_mon+1, T.tm_mday);
  }
  strcpy(date, datep);
}

/*==================================================================
		Coordinate manipulations (trigo)
 *==================================================================*/

#if TRIGO_DEF == 1		/* Add trigonometric definitions */
#undef  TRIGO_DEF
#define TRIGO_DEF 0		/* And now there are included... */

int tr_ou(double o[2], double u[3])
/*++++++++++++++++
.PURPOSE  Compute direction cosines
.RETURNS  0
-----------------*/
{
    u[0] = u[1] = cosd(o[1]) ;
    u[0] *= cosd(o[0]) ;
    u[1] *= sind(o[0]) ;
    u[2] = sind(o[1]) ;
    return(0);
}

int tr_uo(double u[3], double o[2])
/*++++++++++++++++
.PURPOSE  Compute RA+Dec from direction cosines
.RETURNS  0=OK 1=pole -1=undef
-----------------*/
{
  double r2;            /* sqrt(x*x+y*y) */
    r2 = u[0]*u[0] + u[1]*u[1]; o[0] = 0.e0;
    if (r2  == 0.e0) {          /* in case of poles */
        if (u[2] == 0.e0)  { o[0] = o[1] = 0./0.; return(-1); }
        o[1] = ( u[2]>0.e0 ? 90.e0 : -90.e0);
        o[0] = 0.;
        return (1);
    }
    o[1] = atand  ( u[2] / sqrt(r2));
    o[0]  = atan2d (u[1] , u[0] );
    if (o[0] < 0.e0) o[0] += 360.e0;
    return(0);
}

int tr_oR (double o[2], double R[3][3])
/*++++++++++++++++
.PURPOSE  Creates the rotation matrix R[3][3] defined as
	 R[0] (first row) = unit vector towards Zenith
	 R[1] (second row) = unit vector towards East
	 R[2] (third row) = unit vector towards North
.RETURNS 0
-----------------*/
{
    R[2][2] =  cosd(o[1]);
    R[0][2] =  sind(o[1]);
    R[1][1] =  cosd(o[0]);
    R[1][0] =  -sind(o[0]);
    R[1][2] =  0.e0;
    R[0][0] =  R[2][2] * R[1][1];
    R[0][1] = -R[2][2] * R[1][0];
    R[2][0] = -R[0][2] * R[1][1];
    R[2][1] =  R[0][2] * R[1][0];
    return(0);
}

int tr_uu(double u1[3], double u2[3], double R[3][3])
/*++++++++++++++++
.PURPOSE  Rotates the unit vector u1 to u2, as  (u2) = (R) * (u1)
.RETURNS  0
-----------------*/
{
  register int i;
  double u_stack[3];	/* allows same address for input/output      */

    for (i=0; i<3; i++)
	u_stack[i] = R[i][0]*u1[0] +R[i][1]*u1[1] +  R[i][2]*u1[2] ;
    u2[0] = u_stack[0]; 	/* copies to output */
    u2[1] = u_stack[1]; 	/* copies to output */
    u2[2] = u_stack[2]; 	/* copies to output */
    return(0);
}

int tr_uu1(double u1[3], double u2[3], double R[3][3])
/*++++++++++++++++
.PURPOSE  Rotates the unit vector u1 to u2, as  (u2) = (R)^-1^ * (u1)
.RETURNS  0
-----------------*/
{
  register int i;
  double u_stack[3];    /* allows same address for input/output      */
    for (i=0; i<3; i++)
        u_stack[i] = R[0][i]*u1[0] +R[1][i]*u1[1] +  R[2][i]*u1[2] ;
    u2[0] = u_stack[0];         /* copies to output */
    u2[1] = u_stack[1];         /* copies to output */
    u2[2] = u_stack[2];         /* copies to output */
    return(0);
}
#endif
